import * as fs from 'fs';
import * as csvtojson from 'csvtojson';
import { Injectable } from '@nestjs/common';
import { AppGateway } from '../app.gateway';
import { Equipo } from '../equipo/entity/equipo.entity';
import { Institucion } from '../institucion/entity/institucion.entity';
import { InstitucionUsuario } from '../institucion-usuario/entity/institucion-usuario.entity';
import { Operador } from '../operador/entity/operador.entity';
import { CarritoService } from '../carrito/carrito.service';
import { EquipoService } from '../equipo/equipo.service';
import { EquipoProgramaService } from '../equipo-programa/equipo-programa.service';
import { EquipoTipoEntradaService } from '../equipo-tipo-entrada/equipo-tipo-entrada.service';
import { InstitucionService } from '../institucion/institucion.service';
import { InstitucionCarreraService } from '../institucion-carrera/institucion-carrera.service';
import { InstitucionProgramaService } from '../institucion-programa/institucion-programa.service';
import { InstitucionTipoCarritoService } from '../institucion-tipo-carrito/institucion-tipo-carrito.service';
import { InstitucionTipoEntradaService } from '../institucion-tipo-entrada/institucion-tipo-entrada.service';
import { InstitucionUsuarioService } from '../institucion-usuario/institucion-usuario.service';
import { MarcaService } from '../marca/marca.service';
import { ModeloService } from '../modelo/modelo.service';
import { ModuloService } from '../modulo/modulo.service';
import { TipoUsuarioService } from '../tipo-usuario/tipo-usuario.service';
import { UsuarioService } from '../usuario/usuario.service';
import { UploadFileCargaMasivaUsuarioDto } from './dto/upload-file-carga-masiva-usuario.dto';
import { UploadFileCargaMasivaEquipoDto } from './dto/upload-file-carga-masiva-equipo.dto';

@Injectable()
export class UploadFileService {
  constructor(
    private appGateway: AppGateway,
    private carritoService: CarritoService,
    private equipoService: EquipoService,
    private equipoProgramaService: EquipoProgramaService,
    private equipoTipoEntradaService: EquipoTipoEntradaService,
    private institucionService: InstitucionService,
    private institucionCarreraService: InstitucionCarreraService,
    private institucionProgramaService: InstitucionProgramaService,
    private institucionTipoCarritoService: InstitucionTipoCarritoService,
    private institucionTipoEntradaService: InstitucionTipoEntradaService,
    private institucionUsuarioService: InstitucionUsuarioService,
    private marcaService: MarcaService,
    private modeloService: ModeloService,
    private moduloService: ModuloService,
    private tipoUsuarioService: TipoUsuarioService,
    private usuarioService: UsuarioService,
  ) {}

  createEquipos(path: string, admin: Operador): { message: string } {
    let errores: string[] = [];
    let equiposNuevos: Equipo[] = [];

    csvtojson()
      .fromFile(path)
      .then(async (equipos: UploadFileCargaMasivaEquipoDto[]) => {
        for (let i = 0; i < equipos.length; i++) {
          await this.evaluarEquipo(
            admin.institucion,
            i,
            equipos[i],
            errores,
            equiposNuevos,
          );
          if ((i + 1) % 90 === 0 || i + 1 === equipos.length) {
            this.appGateway.equiposNuevos(admin.institucion.id_institucion, {
              errores,
              equiposNuevos,
            });
            errores = [];
            equiposNuevos = [];
          }
        }
        fs.unlink(path, (err) => {});
      });
    return { message: 'Se subió y cargó correctamente su archivo csv.' };
  }

  createUsuarios(path: string, admin: Operador): { message: string } {
    let errores: string[] = [];
    let mensajes: string[] = [];
    let usuariosNuevos: InstitucionUsuario[] = [];

    csvtojson()
      .fromFile(path)
      .then(async (usuarios: UploadFileCargaMasivaUsuarioDto[]) => {
        for (let i = 0; i < usuarios.length; i++) {
          await this.evaluarUsuario(
            admin.institucion,
            i,
            usuarios[i],
            mensajes,
            errores,
            usuariosNuevos,
          );
          if ((i + 1) % 100 === 0 || i + 1 === usuarios.length) {
            this.appGateway.usuariosNuevos(admin.institucion.id_institucion, {
              errores,
              mensajes,
              usuariosNuevos,
            });
            errores = [];
            mensajes = [];
            usuariosNuevos = [];
          }
        }
      });
    return { message: 'Se subió y cargó correctamente tu archivo csv.' };
  }

  downloadLogo(operador: Operador): Promise<string> {
    return this.institucionService
      .findById(operador.institucion.id_institucion)
      .then((institucion) => institucion.logo);
  }

  errorBase(index): string {
    return `Se saltó la linea ${
      index + 2
    } por el siguiente/los siguientes motivo(s):`;
  }

  async evaluarEquipo(
    institucion: Institucion,
    index: number,
    dataEquipo: UploadFileCargaMasivaEquipoDto,
    errores: string[],
    equiposNuevos: Equipo[],
  ) {
    const errorAux = this.errorBase(index);
    let error = errorAux;

    // Si falta agluno de estos campos sacamos error
    if (
      !dataEquipo.carrito ||
      !dataEquipo.entradas ||
      !dataEquipo.equipo ||
      !dataEquipo.marca ||
      !dataEquipo.modelo ||
      !dataEquipo.modulo ||
      !dataEquipo.numero_inventario ||
      !dataEquipo.numero_serie ||
      !dataEquipo.tipo
    ) {
      // Si falta campo carrito, agregamos error a la variable error
      if (!dataEquipo.carrito) error += ' falta del campo carrito';
      // Si falta campo entradas, agregamos error a la variable error
      if (!dataEquipo.entradas) {
        if (error != errorAux) error += ',';
        error += ' falta del campo entradas';
      }
      // Si falta campo equipo, agregamos error a la variable error
      if (!dataEquipo.equipo) {
        if (error != errorAux) error += ',';
        error += ' falta del campo equipo';
      }
      // Si falta campo marca, agregamos error a la variable error
      if (!dataEquipo.marca) {
        if (error != errorAux) error += ',';
        error += ' falta del campo marca';
      }
      // Si falta campo modelo, agregamos error a la variable error
      if (!dataEquipo.modelo) {
        if (error != errorAux) error += ',';
        error += ' falta del campo modelo';
      }
      // Si falta campo modulo , agregamos error a la variable error
      if (!dataEquipo.modulo) {
        if (error != errorAux) error += ',';
        error += ' falta del campo modulo';
      }
      // Si falta campo numero_inventario, agregamos error a la variable error
      if (!dataEquipo.numero_inventario) {
        if (error != errorAux) error += ',';
        error += ' falta del campo numero_inventario';
      }
      // Si falta campo numero_serie, agregamos error a la variable error
      if (!dataEquipo.numero_serie) {
        if (error != errorAux) error += ',';
        error += ' falta del campo numero_serie';
      }
      // Si falta campo tipo, agregamos error a la variable error
      if (!dataEquipo.tipo) {
        if (error != errorAux) error += ',';
        error += ' falta del campo tipo';
      }
      return errores.push(error + '.'); // Agregamos error al array de errores
    }

    const tipoCarrito =
      await this.institucionTipoCarritoService.findTipoCarritoByTipoCarrito(
        dataEquipo.tipo,
      );
    const modulo = await this.moduloService.findModulo(
      institucion,
      dataEquipo.modulo,
    );
    const marca = await this.marcaService.findByMarca(dataEquipo.marca, 'e');
    const modelo = await this.modeloService.findByModelo(
      dataEquipo.modelo,
      'e',
    );
    const carrito =
      tipoCarrito && modulo
        ? await this.carritoService.findCarrito(
            modulo,
            tipoCarrito,
            dataEquipo.carrito,
          )
        : null;

    // Validamos que exista la información mandada del client en la db, si no sacamos error
    if (!tipoCarrito || !marca || !modelo || !modulo || !carrito) {
      // Validamos que existe el tipo de carrito
      if (!tipoCarrito) error += ' no existe este tipo de carrito';
      // Validamos que existe la marca
      if (!marca) {
        if (error != errorAux) error += ',';
        error += ' no existe esta marca';
      }
      // Validamos que existe el modelo
      if (!modelo) {
        if (error != errorAux) error += ',';
        error += ' no existe este modelo';
      }
      // Validamos que existe el módulo
      if (!modulo) {
        if (error != errorAux) error += ',';
        error += ' no existe este módulo';
      }
      // Validamos que existe el carrito
      if (!carrito) {
        if (error != errorAux) error += ',';
        error += ' no existe este carrito';
      }
      return errores.push(error + '.'); // Agregamos el error al array de errores
    }

    // Buscamos un equipo con el sobrenombre enviado del client en el carrito seleccionado
    let equipo = carrito
      ? await this.equipoService.findByEquipo(carrito, dataEquipo.equipo, false)
      : null;

    // Si existe uno sacamos error
    if (equipo)
      return errores.push(
        `${error} ya existe un equipo de cómputo con el sobrenombre: ${equipo.equipo} en el carrito: ${equipo.carrito.carrito}.`,
      );
    // Buscamos un equipo con el numero_inventario enviado del client en la institución del admin
    equipo = await this.equipoService.findByNumeroInventario(
      institucion,
      dataEquipo.numero_inventario,
      false,
    );
    // Si existe uno saca error
    if (equipo)
      return errores.push(
        `${error} ya existe un equipo con ese número de inventario en esta institución.`,
      );
    // Creamos y guardamos un registro
    await this.equipoService
      .create(
        dataEquipo.equipo,
        carrito,
        dataEquipo.numero_inventario,
        dataEquipo.numero_serie,
        marca,
        modelo,
      )
      .then(async (equipoNuevo) => {
        const entradas = dataEquipo.entradas.split(',');
        const programas = dataEquipo.programas
          ? dataEquipo.programas.split(',')
          : [];

        // Inicializamos los arrays
        equipoNuevo.programas = [];
        equipoNuevo.tiposEntradas = [];
        // Recorremos el array de entradas
        for (let j = 0; j < entradas.length; j++) {
          const tipoEntrada =
            await this.institucionTipoEntradaService.findTipoEntradaByTipoEntrada(
              entradas[j].trim(),
            );
          const existeEquipoTipoEntrada = tipoEntrada
            ? await this.equipoTipoEntradaService.findByEquipoTipoEntrada(
                equipoNuevo,
                tipoEntrada,
              )
            : null;

          if (!tipoEntrada || existeEquipoTipoEntrada) {
            // Validamos que exista el tipo de entrada que se quiere asignar al equipo
            if (!tipoEntrada)
              errores.push(
                `No se asignó el conector: ${
                  entradas[j]
                }, al equipo de la linea ${
                  index + 2
                } porque no existe este conector.`,
              );
            // Validamos que el tipo de entrada no este ya asignado al equipo
            if (existeEquipoTipoEntrada)
              errores.push(
                `El conector: ${
                  entradas[j]
                }, ya fue asignado al equipo de la linea ${index + 2}.`,
              );
            // Nos saltamos esta entrada
            continue;
          }
          // Asignamos el tipo de entrada al equipo
          await this.equipoTipoEntradaService
            .create(equipoNuevo, tipoEntrada)
            .then((equipoTipoEntrada) =>
              equipoNuevo.tiposEntradas.push(equipoTipoEntrada),
            );
        }
        // Recorremos el array de programas
        for (let j = 0; j < programas.length; j++) {
          const programa =
            await this.institucionProgramaService.findProgramaByPrograma(
              programas[j].trim(),
            );
          const existeEquipoPrograma = programa
            ? await this.equipoProgramaService.findByEquipoPrograma(
                equipoNuevo,
                programa,
              )
            : null;

          if (!programa || existeEquipoPrograma) {
            // Validamos que exista el programa que se quiere asignar al equipo
            if (!programa)
              errores.push(
                `No se asignó el software: ${
                  programas[j]
                }, al equipo de la linea ${
                  index + 2
                } porque no existe este software.`,
              );
            // Validamos que el programa no este ya asignado al equipo
            if (existeEquipoPrograma)
              errores.push(
                `El programa: ${
                  programas[j]
                }, ya lo tiene asignado el equipo de la linea ${index + 2}.`,
              );
            // Nos saltamos este programa
            continue;
          }
          // Asignamos el programa al equipo
          await this.equipoProgramaService
            .create(equipoNuevo, programa)
            .then((equipoPrograma) =>
              equipoNuevo.programas.push(equipoPrograma),
            );
        }
        // Agregamos el equipo nuevo al array de equipos nuevos
        equiposNuevos.push(equipoNuevo);
      });
  }

  async evaluarUsuario(
    institucion: Institucion,
    index: number,
    dataUsuario: UploadFileCargaMasivaUsuarioDto,
    mensajes: string[],
    errores: string[],
    usuariosNuevos: InstitucionUsuario[],
  ) {
    const errorAux = this.errorBase(index);
    let error = errorAux;

    // Si falta agluno de estos campos sacamos error
    if (!dataUsuario.numero_cuenta || !dataUsuario.tipo_usuario) {
      // Si falta campo numero_cuenta, agregamos error a la variable error
      if (!dataUsuario.numero_cuenta) error += ' falta del campo numero_cuenta';
      // Si falta campo tipo_usuario, agregamos error a la variable error
      if (!dataUsuario.tipo_usuario) {
        if (error != errorAux) error += ',';
        error += ' falta del campo tipo_usuario';
      }
      return errores.push(`${error}.`);
    }

    const tipoUsuario = await this.tipoUsuarioService.findByTipoUsuario(
      dataUsuario.tipo_usuario,
    );
    const usuario = await this.usuarioService.findByUsuario(
      dataUsuario.numero_cuenta,
    );
    const carrera =
      tipoUsuario && (tipoUsuario.id_tipo_usuario === 5 || dataUsuario.carrera)
        ? await this.institucionCarreraService.findCarreraByCarrera(
            tipoUsuario.id_tipo_usuario === 5
              ? 'PROFESOR'
              : dataUsuario.carrera,
            false,
          )
        : null;
    const institucionCarrera = carrera
      ? await this.institucionCarreraService.findByIdInstitucionIdCarrera(
          institucion,
          carrera,
        )
      : null;

    // Validamos que exista la información mandada del client en la db, si no sacamos error
    if (
      !tipoUsuario ||
      tipoUsuario.id_tipo_usuario < 5 ||
      !carrera ||
      (carrera && !institucionCarrera)
    ) {
      // Validamos que existe el tipo de usuario
      if (!tipoUsuario) error += ' no existe este tipo de usuario.';
      // Validamos que el id tipo usuario no sea menor a 5
      else if (tipoUsuario.id_tipo_usuario < 5)
        ' no se puede asignar este tipo de usuario a este usuario.';
      // Validamos que existe la carrera
      if (!carrera) error += ' falta del campo carrera para el alumno.';
      // Validamos que la carrera enviada del client pertenezca a la institución del admin
      else if (!institucionCarrera)
        error += ' la carrera no pertenece a tu institución.';
      return errores.push(error);
    }

    // Solución a error de encriptación de caracteres en nombres
    let nombre = `${dataUsuario.apellido_p.trim()} ${dataUsuario.apellido_m.trim()} ${dataUsuario.nombres.trim()}`;

    while (nombre.search('√ë') != -1) {
      nombre = nombre.replace('√ë', 'Ñ');
    }
    while (nombre.search('√É∆í√¢‚Ç¨Àú') != -1) {
      nombre = nombre.replace('√É∆í√¢‚Ç¨Àú', 'Ñ');
    }
    // Si ya existe el usuario
    if (usuario) {
      // Buscamos si el usuario ya tiene asignada la carrera
      let existeInstitucionUsuario =
        await this.institucionUsuarioService.findByIdUsuarioIdInstitucionCarrera(
          usuario,
          institucionCarrera,
        );

      // Si si ya la tiene asignada y no esta activa
      if (existeInstitucionUsuario && !existeInstitucionUsuario.activo) {
        // La activamos
        existeInstitucionUsuario.activo = true;
        await this.institucionUsuarioService
          .save(existeInstitucionUsuario)
          .then((institucionUsuario) =>
            mensajes.push(
              `Se activo la carrera: ${institucionUsuario.institucionCarrera.carrera.carrera} del usuario: ${institucionUsuario.usuario.usuario}.`,
            ),
          );
      }
      // Si no la tiene asignada
      else if (!existeInstitucionUsuario)
        // Se la asignamos
        await this.institucionUsuarioService
          .create(institucionCarrera, usuario)
          .then((institucionUsuario) =>
            mensajes.push(
              `Se asignó la carrera: ${institucionUsuario.institucionCarrera.carrera.carrera} al usuario: ${institucionUsuario.usuario.usuario}.`,
            ),
          );
      // Si el nombre, tipo de usuario o correo enviado es distinto al que se tiene en la db se actualiza
      if (
        usuario.nombre != nombre ||
        usuario.tipoUsuario.id_tipo_usuario != tipoUsuario.id_tipo_usuario ||
        (dataUsuario.correo && usuario.correo != dataUsuario.correo)
      ) {
        // Si el id tipo usuario es distinto al que se tiene en base se actualiza
        if (usuario.tipoUsuario.id_tipo_usuario != tipoUsuario.id_tipo_usuario)
          usuario.tipoUsuario = tipoUsuario;
        // Si el nombre es distinto al que se tiene en base se actualiza
        if (usuario.nombre != nombre) usuario.nombre = nombre;
        // Si se mando correo y el correo es distinto al que se tiene en base se actualiza
        if (dataUsuario.correo && usuario.correo != dataUsuario.correo)
          usuario.correo = dataUsuario.correo;
        // Guardamos cambios de usuario
        await this.usuarioService
          .save(usuario)
          .then((_) =>
            mensajes.push(
              `Se actualizaron los datos de usuario: ${usuario.usuario}`,
            ),
          );
      }
    }
    // Si no existe el usuario
    else {
      // Si falta agluno de estos campos sacamos error
      if (
        !dataUsuario.nombres ||
        !dataUsuario.apellido_m ||
        !dataUsuario.apellido_p ||
        !dataUsuario.correo ||
        (tipoUsuario.id_tipo_usuario === 5 && !dataUsuario.rfc)
      ) {
        // Si falta campo rfc y el tipo de usuario es profesor, agregamos error a la variable error
        if (tipoUsuario.id_tipo_usuario === 5 && !dataUsuario.rfc)
          error += ' falta del campo rfc';
        // Si falta campo nombres, agregamos error a la variable error
        if (!dataUsuario.nombres) {
          if (error != errorAux) error += ',';
          error += ' falta del campo nombres';
        }
        // Si falta campo apellido_m, agregamos error a la variable error
        if (!dataUsuario.apellido_m) {
          if (error != errorAux) error += ',';
          error += ' falta del campo apellido_m';
        }
        // Si falta campo apellido_p, agregamos error a la variable error
        if (!dataUsuario.apellido_p) {
          if (error != errorAux) error += ',';
          error += ' falta del campo apellido_p';
        }
        // Si falta campo correo, agregamos error a la variable error
        if (!dataUsuario.correo) {
          if (error != errorAux) error += ',';
          error += ' falta del campo correo';
        }
        return errores.push(`${error}.`); // Agregamos error al array de errores
      }
      // Creamos un usuario y le asignamos la carrera enviada
      await this.usuarioService
        .create(
          dataUsuario.numero_cuenta,
          nombre,
          tipoUsuario,
          institucionCarrera,
          dataUsuario.correo,
          dataUsuario.rfc ? dataUsuario.rfc : '',
        )
        .then((institucionUsuario) => usuariosNuevos.push(institucionUsuario));
    }
  }

  uploadLogo(
    file: Express.Multer.File,
    admin: Operador,
  ): Promise<{ message: string }> {
    return this.institucionService
      .findById(admin.institucion.id_institucion)
      .then((institucion) => {
        // Si ya hay registro de un logo eliminamos el archivo
        if (institucion.logo) fs.unlink(institucion.logo, (err) => {});
        // Guardamos la nueva ubicación del logo
        institucion.logo = `${file.destination}/${file.filename}`;
        // Guardamos los cambios
        return this.institucionService
          .save(institucion)
          .then((_) => ({ message: 'Se subió correctamente el logo.' }));
      });
  }
}
